@@ -16,6 +16,9 @@
   Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
   MA 02139, USA.
 */
+
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -35,11 +38,13 @@
 #include <sys/time.h>
 #include <fcntl.h>
 #include <errno.h>
+#include <libgen.h>
 
 /* Local includes */
 #include "xvm.h"
 #include "simple_auth.h"
 #include "mcast.h"
+#include "tcp_listener.h"
 #include "options.h"
 
 #define SCHEMA_COMPAT '\xfe'
@@ -64,14 +69,6 @@ assign_debug(fence_virt_args_t *args, struct arg_info *arg, char *value)
 }
 
 
-static inline void
-assign_foreground(fence_virt_args_t *args, struct arg_info *arg,
-		  char *value)
-{
-	args->flags |= F_FOREGROUND;
-}
-
-
 static inline void
 assign_family(fence_virt_args_t *args, struct arg_info *arg,
 	      char *value)
@@ -103,6 +100,16 @@ assign_address(fence_virt_args_t *args, struct arg_info *arg, char *value)
 	args->net.addr = strdup(value);
 }
 
+static inline void
+assign_ip_address(fence_virt_args_t *args, struct arg_info *arg, char *value)
+{
+	if (!value)
+		return;
+
+	if (args->net.ipaddr)
+		free(args->net.ipaddr);
+	args->net.ipaddr = strdup(value);
+}
 
 static inline void
 assign_channel_address(fence_virt_args_t *args, struct arg_info *arg, char *value)
@@ -119,38 +126,72 @@ assign_channel_address(fence_virt_args_t *args, struct arg_info *arg, char *valu
 static inline void
 assign_port(fence_virt_args_t *args, struct arg_info *arg, char *value)
 {
+	char *p;
+	int ret;
+
 	if (!value)
 		return;
 
-	args->net.port = atoi(value);
-	if (args->net.port <= 0 || args->net.port >= 65500) {
+	ret = strtol(value, &p, 0);
+	if (ret <= 0 || ret >= 65536 || *p != '\0') {
 		printf("Invalid port: '%s'\n", value);
 		args->flags |= F_ERR;
-	}
+	} else
+		args->net.port = ret;
 }
 
+static inline void
+assign_cid(fence_virt_args_t *args, struct arg_info *arg, char *value)
+{
+	char *p;
+	unsigned long ret;
+
+	if (!value) {
+		args->net.cid = 2;
+		return;
+	}
+
+	ret = strtoul(value, &p, 0);
+	if (!p || *p != '\0' || ret < 2 || ret >= 0xffffffff) {
+		printf("Invalid CID: '%s'\n", value);
+		args->flags |= F_ERR;
+	} else
+		args->net.cid = ret;
+}
 
 static inline void
 assign_interface(fence_virt_args_t *args, struct arg_info *arg, char *value)
 {
+	int ret;
+
 	if (!value)
 		return;
 
-	args->net.ifindex = if_nametoindex(value);
+	ret = if_nametoindex(value);
+	if (ret <= 0) {
+		printf("Invalid interface: %s: %s\n", value, strerror(errno));
+		args->net.ifindex = 0;
+	}
+
+	args->net.ifindex = ret;
 }
 
 
 static inline void
 assign_retrans(fence_virt_args_t *args, struct arg_info *arg, char *value)
 {
+	char *p;
+	int ret;
+
 	if (!value)
 		return;
 
-	args->retr_time = atoi(value);
-	if (args->retr_time <= 0) {
+	ret = strtol(value, &p, 0);
+	if (ret <= 0 || *p != '\0') {
 		printf("Invalid retransmit time: '%s'\n", value);
 		args->flags |= F_ERR;
-	}
+	} else
+		args->retr_time = ret;
 }
 
 static inline void
@@ -232,10 +273,12 @@ assign_op(fence_virt_args_t *args, struct arg_info *arg, char *value)
 		args->op = FENCE_STATUS;
 	} else if (!strcasecmp(value, "monitor")) {
 		args->op = FENCE_DEVSTATUS;
-	} else if (!strcasecmp(value, "list")) {
+	} else if (!strcasecmp(value, "list") || !strcasecmp(value, "list-status")) {
 		args->op = FENCE_HOSTLIST;
 	} else if (!strcasecmp(value, "metadata")) {
 		args->op = FENCE_METADATA;
+	} else if (!strcasecmp(value, "validate-all")) {
+		args->op = FENCE_VALIDATEALL;
 	} else {
 		printf("Unsupported operation: %s\n", value);
 		args->flags |= F_ERR;
@@ -308,16 +351,36 @@ assign_uuid_lookup(fence_virt_args_t *args, struct arg_info *arg, char *value)
 static inline void
 assign_timeout(fence_virt_args_t *args, struct arg_info *arg, char *value)
 {
+	char *p;
+	int ret;
+
 	if (!value)
 		return;
 
-	args->timeout = atoi(value);
-	if (args->timeout <= 0) {
+	ret = strtol(value, &p, 0);
+	if (ret <= 0 || *p != '\0') {
 		printf("Invalid timeout: '%s'\n", value);
 		args->flags |= F_ERR;
-	}
+	} else
+		args->timeout = ret;
 }
 
+static inline void
+assign_delay(fence_virt_args_t *args, struct arg_info *arg, char *value)
+{
+	char *p;
+	int ret;
+
+	if (!value)
+		return;
+
+	ret = strtol(value, &p, 0);
+	if (ret < 0 || *p != '\0') {
+		printf("Invalid delay: '%s'\n", value);
+		args->flags |= F_ERR;
+	} else
+		args->delay = ret;
+}
 
 static inline void
 assign_help(fence_virt_args_t *args, struct arg_info *arg, char *value)
@@ -333,19 +396,6 @@ assign_version(fence_virt_args_t *args, struct arg_info *arg, char *value)
 }
 
 
-static inline void
-assign_uri(fence_virt_args_t *args, struct arg_info *arg, char *value)
-{
-#if 0
-	if (args->uri)
-		free(args->uri);
-
-	/* XXX NO validation yet */
-	args->uri = strdup(value);
-#endif
-}
-
-
 static void
 print_desc_xml(const char *desc)
 {
@@ -398,6 +448,16 @@ static struct arg_info _arg_info[] = {
 	  "Multicast address (default=" IPV4_MCAST_DEFAULT " / " IPV6_MCAST_DEFAULT ")",
 	  assign_address },
 
+	{ 'T', "-T <address>", "ipaddr",
+          0, "string", "127.0.0.1",
+	  "IP address to connect to in TCP mode (default=" IPV4_TCP_ADDR_DEFAULT " / " IPV6_TCP_ADDR_DEFAULT ")",
+	  assign_ip_address },
+
+	{ 'S', "-S <cid>", "vsock",
+          0, "int", "2",
+	  "vm socket CID to connect to in vsock mode",
+	  assign_cid },
+
 	{ 'A', "-A <address>", "channel_address",
           0, "string", "10.0.2.179",
 	  "VM Channel IP address (default=" DEFAULT_CHANNEL_IP ")",
@@ -405,7 +465,7 @@ static struct arg_info _arg_info[] = {
 
 	{ 'p', "-p <port>", "ipport",
           0, "string", "1229",
-	  "Multicast or VMChannel IP port (default=1229)",
+	  "TCP, Multicast, VMChannel, or VM socket port (default=1229)",
 	  assign_port },
 
 	{ 'I', "-I <interface>", "interface",
@@ -446,12 +506,12 @@ static struct arg_info _arg_info[] = {
 	{ '\xff', NULL, "option",
 	  /* Deprecated */
 	  0, "string", "reboot",
-	  "Fencing option (null, off, on, [reboot], status, list, monitor, metadata)",
+	  "Fencing option (null, off, on, [reboot], status, list, list-status, monitor, validate-all, metadata)",
 	  assign_op },
 
 	{ 'o', "-o <operation>", "action",
 	  0, "string", "reboot",
-	  "Fencing action (null, off, on, [reboot], status, list, monitor, metadata)",
+	  "Fencing action (null, off, on, [reboot], status, list, list-status, monitor, validate-all, metadata)",
 	  assign_op },
 
 	{ 'H', "-H <domain>", "port",
@@ -484,11 +544,11 @@ static struct arg_info _arg_info[] = {
  	  "Help (alternate)", 
 	  assign_help },
 
-	{ 'U', "-U", "uri",
-	  0, "boolean", "qemu:///system",
-	  "URI for Hypervisor (default: auto detect)",
-	  assign_uri },
-	  
+	{ 'w', "-w <delay>", "delay",
+	  0, "string", "0",
+	  "Fencing delay (in seconds; default=0)",
+	  assign_delay },
+
 	{ 'V', "-V", NULL,
 	  0, "boolean", "0",
  	  "Display version and exit", 
@@ -546,9 +606,11 @@ args_init(fence_virt_args_t *args)
 	args->net.hash = DEFAULT_HASH;
 	args->net.auth = DEFAULT_AUTH;
 	args->net.addr = NULL;
+	args->net.ipaddr = NULL;
+	args->net.cid = 0;
 	args->net.port = DEFAULT_MCAST_PORT;
 	args->net.ifindex = 0;
-	args->net.family = PF_INET;
+	args->net.family = 0;  /* auto */
 	args->serial.device = NULL;
 	args->serial.speed = strdup(DEFAULT_SERIAL_SPEED);
 	args->serial.address = strdup(DEFAULT_CHANNEL_IP);
@@ -556,6 +618,7 @@ args_init(fence_virt_args_t *args)
 	args->retr_time = 20;
 	args->flags = 0;
 	args->debug = 0;
+	args->delay = 0;
 }
 
 
@@ -574,18 +637,26 @@ args_print(fence_virt_args_t *args)
 	printf("-- args @ %p --\n", args);
 	_pr_str(args->domain);
 	_pr_int(args->op);
-
-	_pr_str(args->net.key_file);
-	_pr_int(args->net.hash);
-	_pr_str(args->net.addr);
-	_pr_int(args->net.auth);
-	_pr_int(args->net.port);
-	_pr_int(args->net.ifindex);
-	_pr_int(args->net.family);
+	_pr_int(args->mode);
+	_pr_int(args->debug);
 	_pr_int(args->timeout);
+	_pr_int(args->delay);
 	_pr_int(args->retr_time);
 	_pr_int(args->flags);
-	_pr_int(args->debug);
+
+	_pr_str(args->net.addr);
+	_pr_str(args->net.ipaddr);
+	_pr_int(args->net.cid);
+	_pr_str(args->net.key_file);
+	_pr_int(args->net.port);
+	_pr_int(args->net.hash);
+	_pr_int(args->net.auth);
+	_pr_int(args->net.family);
+	_pr_int(args->net.ifindex);
+
+	_pr_str(args->serial.device);
+	_pr_str(args->serial.speed);
+	_pr_str(args->serial.address);
 	printf("-- end args --\n");
 }
 
@@ -622,7 +693,7 @@ output_help_text(int arg_width, int help_width, const char *arg, const char *des
 	int len;
 
 	memset(out_buf, 0, sizeof(out_buf));
-	strncpy(out_buf, desc, sizeof(out_buf));
+	strncpy(out_buf, desc, sizeof(out_buf) - 1);
 	start = out_buf;
 
 	do {
@@ -742,6 +813,8 @@ args_metadata(char *progname, const char *optstr)
 	printf("\t<action name=\"status\" />\n");	
 	printf("\t<action name=\"monitor\" />\n");	
 	printf("\t<action name=\"list\" />\n");	
+	printf("\t<action name=\"list-status\" />\n");	
+	printf("\t<action name=\"validate-all\" />\n");	
 	printf("</actions>\n");
 	printf("</resource-agent>\n");
 }
@@ -835,6 +908,7 @@ args_get_stdin(const char *optstr, fence_virt_args_t *args)
 				"parse warning: "
 				"illegal variable '%s' on line %d\n", name,
 				line);
+			args->flags |= F_ERR;
 			continue;
 		}
 
